# This collection of functions is intended to provide an S-language
# level implementation of converters between R and Java of list
# objects.  By doing this at this level, we pay a performance penalty
# because we first of all have to search through the converters, and
# also the conversion is done in interpreted code and not C code.

convertListToJava =
function(x, ...)
{
  if(length(names(x)) == 0) {
        # create an array in Java and convert each element.
     val = .JavaArrayConstructor("java.lang.Object", dim = length(x))
     for(i in seq(length = length(x))) {
       .JavaSetArrayElement(val, x[[i]], i, .convert = NULL)
     }
  } else {
     val = .JNew("org.omegahat.Environment.Utils.OrderedTable", length(x))
     for(i in names(x)) {
       val$put(i, x[[i]])
     }
  }

  return(val)
}  

convertNamedListToJava <-
    function(x, ...)
{
    h <- .JNew("java.util.HashMap")
    for(n in names(x)) {
        .Java(h, "put", n, x[[n]])
    }
    h
}

convertOrderedTableFromJava =
function(x, ...)
{
  ans = vector("list")
  p = x$keys()  # A java.util.Enumeration object.
      # Many calls but little memory per-call.  
  while(p$hasMoreElements()) {
    i = p$nextElement()
    ans[[i]] = x$get(i)
  }

  ans
}



convertOrderedTableFromJava =
function(x, ...)
{
  ans = vector("list")
  for(i in x$orderedKeys()) {
    ans[[i]] = x$get(i)
  }

  ans
}  


convertVectorFromJava =
function(x, ...)
{
  n = x$size()

  ans = vector("list", n)
  for(i in seq(length = n)) {
    ans[[i]] = x$elementAt(as.integer(i-1))
  }

  simplifyListToVector(ans)
}


simplifyListToVector =
  # 
  # Taken from sapply()
  #
function(x)
{
    if(length(x) &&
       length(common.len <- unique(unlist(lapply(x, length)))) == 1)
    {
        types = unique(sapply(x, class))
        if(length(types) == 1 || sort(types) == c("integer", "numeric")) {
            if (common.len == 1) 
                x <- unlist(x, recursive = FALSE)
            else if (common.len > 1) 
                x <- array(unlist(x, recursive = FALSE),
                           dim = c(common.len, length(x)),
                           dimnames = if (!(is.null(n1 <- names(x[[1]])) &
                                            is.null(n2 <- names(x))))
                           list(n1, n2))
        }
    }
    x
}


convertArrayFromJava <-
function(x, ...)
{
   n = .JavaArrayLength(x)

   a = vector("list", n)

   for(i in seq(length = n)) {
        a[[i]] = x[i]
   }

   simplifyListToVector(a)
}  
