/* #line 99 "Converters.cweb" */
#ifndef CONVERTERS_JAVA_H
#define CONVERTERS_JAVA_H

#include <jni.h>
#include "RSCommon.h"

#if defined(_R_)
#include "Rinternals.h"
#endif

#include "CtoJava.h"


/* #line 126 "Converters.cweb" */
typedef struct _RSFromJavaConverter RSFromJavaConverter;

typedef jboolean (*FromJavaConverterMatch)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

typedef USER_OBJECT_ (*FromJavaConverter)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

struct _RSFromJavaConverter {
   /* Determines whether this element is to be used for
      converting the specified object. If this returns
      false, it is skipped. Otherwise, it is used to
      perform the conversion.
    */
 FromJavaConverterMatch match;

   /* The method that performs the computation
      to convert the Java object to an R/S object.
    */

 FromJavaConverter converter;

   /* Handle arrays automaticall by matching on the first
      element and converting the array element wise using this
      elements converter.
      If this is false, then the match and converter routines
      can and must handle arrays differently.
    */
 jboolean autoArray;

    /*
      Available to the match and converter routines for storing
      local data.
     */
 void *userData;

 const char *description;


     /* Next element in the linked list of converters. */
 RSFromJavaConverter *next;
};

extern RSFromJavaConverter *FromJavaConverters;




typedef struct _RSToJavaConverter RSToJavaConverter;

typedef jboolean (*ToJavaConverterMatch)(USER_OBJECT_ obj, jclass type, java_type, const char *signature, jboolean isArray, int arrayLen, JNIEnv *,  RSToJavaConverter *);
typedef jobject (*ToJavaConverter)(USER_OBJECT_ obj, jclass type, java_type, const char *signature, jboolean isArray, int arrayLen, JNIEnv *,  RSToJavaConverter *);

struct _RSToJavaConverter {

  ToJavaConverterMatch match;
  ToJavaConverter      converter;
  jboolean             autoArray;
  void                *userData;

  const char *description;

  RSToJavaConverter *next;
};

extern RSToJavaConverter *ToJavaConverters;



/* #line 113 "Converters.cweb" */
#ifdef __cplusplus
extern "C" {
#endif
/* #line 194 "Converters.cweb" */
USER_OBJECT_
userLevelJavaConversion(JNIEnv *env, jobject obj, const char *dataType, jboolean isArray, int arrayLength, int *ok);

/* #line 313 "Converters.cweb" */
RSFromJavaConverter *removeFromJavaConverterByIndex(int which);
RSFromJavaConverter *removeFromJavaConverterByDescription(const char *desc, int *which);

/* #line 447 "Converters.cweb" */
jboolean
SimpleExactClassMatch(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);
jboolean
InstanceOfFromClassMatch(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);
jboolean
AssignableFromClassMatch(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

typedef enum {InstanceOf=1, Equals, AssignableFrom} JavaConverterMatch;

/* #line 514 "Converters.cweb" */
/*
 Routine that establishes the basic converters for
 Method and Constructor, the reference classes and
 also any that convert from R to Java.
*/
void RS_JAVA(registerDefaultConverters)(JNIEnv *env);

/* #line 672 "Converters.cweb" */
/*
  High-level mechanism for registering a converter for Java to R objects.
 */
RSFromJavaConverter *
addFromJavaConverterInfo(FromJavaConverterMatch match, FromJavaConverter converter, jboolean autoArray, void *userData, const char *description, int *index);

/* #line 702 "Converters.cweb" */
RSToJavaConverter * addToJavaConverterInfo(ToJavaConverterMatch match, ToJavaConverter converter, jboolean autoArray, void *userData, const char *description, int *index);
int addToJavaConverter(RSToJavaConverter *cvt, int pos);


/* #line 765 "Converters.cweb" */
int
getNumConverters(int which);

/* #line 854 "Converters.cweb" */
const char *fromJavaConverterDescription(RSFromJavaConverter *cvt);

/* #line 904 "Converters.cweb" */
/*
 Can be used by a converter that discovers it is not capable
 of converting a Java object to an R object.
 This calls the assign method in the Omega Interface Manager
 which stores the value in the anonymous reference database
 and returns an AnonymousReference object identifying it.
 */
USER_OBJECT_ anonymousAssign(JNIEnv *env, jobject obj, jclass klass);

/* #line 956 "Converters.cweb" */
const char *initAssignMethod(JNIEnv *env);

/* #line 973 "Converters.cweb" */
/*
 Routine to traverse the list of converters from R to Java and check
 whether any of them applies to the object being converted and the target class.
 The first that does is applied and expected to create and return a Java object.
 */
jobject userLevelToJavaConversion(JNIEnv *env, USER_OBJECT_ obj, jclass klass, java_type which, const char *sig, int arrayLen, jboolean isArray, int *done);

/* #line 1014 "Converters.cweb" */
/*
 The basic matching for determining whether a converter applies
 for turning an R object to Java. This compares the name of the
 target class, given by the argument signature, and the name
 stored in the converter's userData field.
 */
jboolean SimpleRTargetMatch(USER_OBJECT_ obj, jclass targetClass, java_type type, char *signature, jboolean isArray, int arrayLen, JNIEnv *env,  RSToJavaConverter *This);

/* #line 1039 "Converters.cweb" */
USER_OBJECT_ RS_JAVA(ReferenceConverter)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 1051 "Converters.cweb" */
jboolean
RS_JAVA(ReferenceClassMatch)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 1074 "Converters.cweb" */
USER_OBJECT_
RS_JAVA(PropertyConverter)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 1204 "Converters.cweb" */
USER_OBJECT_ RS_JAVA(convertIntegerArray)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 1263 "Converters.cweb" */
USER_OBJECT_ RS_JAVA(convertDoubleArray)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 1294 "Converters.cweb" */
USER_OBJECT_
RS_JAVA(convertBooleanArray)(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converter);

/* #line 117 "Converters.cweb" */
#ifdef __cplusplus
}
#endif

#endif

