#include "ROmegahatInterpreter.h"
#include <stdlib.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

#include "RSCommon.h"
#include <Rinternals.h>
#include <Rembedded.h>

#include <Rversion.h>

#ifndef Win32

#if (R_VERSION >= R_Version(2,3,0))
#define CSTACK_DEFNS 1
#include <Rinterface.h>
#endif /* R_VERSION > 2.3.0 */

#else

#if (R_VERSION >= R_Version(2,3,0))
#if !defined(HAVE_UINTPTR_T) && !defined(uintptr_t)
    typedef unsigned long uintptr_t;
#endif
    extern uintptr_t R_CStackLimit;
#endif /* R_VERSION >= 2.3.0 */

#endif /* Win32 */

    extern Rboolean R_Interactive;
    void setup_Rmainloop(void);
    void run_Rmainloop(void);

#include "CtoJava.h"
#include "RStoJava.h"

#include "Converters.h"

#ifdef __cplusplus
}
#endif

#define VMENV (*env)->

/*
  The implementation of the initialization of an embedded R from Java.
*/
JNIEXPORT jboolean JNICALL
Java_org_omegahat_R_Java_ROmegahatInterpreter_initR(JNIEnv *env, jobject engine, jobjectArray args)
{
    int argc = VMENV GetArrayLength(env, args);
    char **argv;
    jstring jtmp;
    const char *tmp;
    int i;

    argv = (char **) malloc(sizeof(char*) * argc);

    for(i = 0 ; i < argc; i++) {
        jtmp = (jstring) VMENV GetObjectArrayElement(env, args, i);
        tmp = VMENV GetStringUTFChars(env, jtmp, NULL);
        argv[i] = strdup(tmp);
        VMENV ReleaseStringUTFChars(env, jtmp, tmp);
    }

    int stat = Rf_initialize_R(argc, argv);
    free(argv);
    if (stat != 0) {
        REprintf("failed to initialize R");
        return JNI_FALSE;
    }
#if (R_VERSION >= R_Version(2,3,0))
    R_CStackLimit = (uintptr_t) -1;
#endif
    R_Interactive = 1;
    setup_Rmainloop();

    return JNI_TRUE;
}



/*
  This is the simple non-terminating event loop
  for R that process the different input sources
  such as the command line and also X11 graphics device, etc.
*/
JNIEXPORT void JNICALL
Java_org_omegahat_R_Java_ROmegahatInterpreter_run(JNIEnv *env, jobject engine)
{
    /*
       If we call this here
       setup_Rmainloop();
       we replicate the banner message.
    */
    run_Rmainloop();
}

JNIEXPORT jboolean JNICALL
Java_org_omegahat_R_Java_ROmegahatInterpreter_registerNativeConverters(JNIEnv *env, jobject manager)
{
    RS_JAVA(registerDefaultConverters(getJNIEnv()));
    return JNI_TRUE;
}

JNIEXPORT jboolean JNICALL
Java_org_omegahat_R_Java_ROmegahatInterpreter_endR(JNIEnv *env, jobject engine)
{
    Rf_endEmbeddedR(0);
    return JNI_TRUE;
}

#ifdef DEFINE_CALL
int eval_R_command(const char *funcName);

JNIEXPORT void JNICALL
Java_RJava_Rengine_callFunc(JNIEnv *env, jobject engine, jstring func)
{
    const char *funcName = VMENV GetStringUTFChars(env, func, NULL);
/* setup_Rmainloop(); */
    eval_R_command(funcName);

    run_Rmainloop();
    VMENV ReleaseStringUTFChars(env, func, funcName);

}

/*
  This is a simple version of a function call that takes
  a single argument which is constructed as `1:10'.
  It allows us to test different function calls such
  as plot(1:10), print(1:10), sum(1:10), etc.
*/
int
eval_R_command(const char *funcName)
{
    SEXP e;
    SEXP fun;
    SEXP arg;
    int i;

    fun = Rf_findFun(Rf_install((char *)funcName),  R_GlobalEnv);
    PROTECT(fun);
    PROTECT(arg = NEW_INTEGER(10));
    for(i = 0; i < GET_LENGTH(arg); i++)
        INTEGER_DATA(arg)[i]  = i + 1;

    e = allocVector(LANGSXP, 2);
    PROTECT(e);
    SETCAR(e, fun);
    SETCAR(CDR(e), arg);

    /* Evaluate the call to the R function.
       Ignore the return value.
    */
    eval(e, R_GlobalEnv);
    UNPROTECT(3);
    return 0;
}

#endif /* DEFINE_CALL */
