/* #line 1843 "RStoJava.cweb" */
#if defined(_R_)
#include "RtoJava.h"
#else
#include "StoJava.h"
#endif

#include "RStoJava.h"
#include "JNIUtils.h"
#include "ManagerInterfaceUtils.h"

#include "Converters.h"
#include "ErrorHandling.h"

#include "References.h"
#include "Reflectance.h"        /* getClassName */

#include "ConverterExamples.h" /* For ConverterFunction */

/* #line 1097 "RStoJava.cweb" */
const int FOREIGN_REFERENCE_ID_SLOT = 1;
const int FOREIGN_REFERENCE_CLASSNAME_SLOT = 2;
const int FOREIGN_REFERENCE_CLASS_SLOT = 3;

/* #line 784 "RStoJava.cweb" */
void *
s_to_java_basic(JNIEnv *env, USER_OBJECT_ obj, const char *sig,
                const char **actual_sig_ptr)
{
    /* this code is intended only for basic objects; it ignores classes,
       expecting the calling S or C code to have determined that it's
       proper to just convert the basic data in the object */
    jboolean isArray, isBuiltIn, isObject;
    long n;
    void *value;
    enum java_types which;
    jclass klass = NULL_JAVA_OBJECT;

    jboolean isReference = JNI_FALSE;

    if(obj == NULL_USER_OBJECT)
        return NULL_JAVA_OBJECT;

    if(sig == (char*) NULL || sig[0] == '\0') {
        sig = getJavaSignature(obj, &isReference);
#if LDEBUG
        printf("Retrieved signature %s\n",sig);  fflush(stdout);
#endif
        if(!sig) {
            PROBLEM "No default Java type for S class ???"
                ERROR;
            return NULL_JAVA_OBJECT;
        }
    }

#if LDEBUG
    printf("[s_to_java_basic] (sig) %s \n", sig);
    fflush(stdout);
#endif

    which = which_java_type(sig, &isArray, &isBuiltIn, &isObject);

#if LDEBUG
    printf("[s_to_java_basic] (type) %d (reference) %d, built-in type %s,  array %s\n",
           which, isReference, isBuiltIn ? "yes" : "no",
           isArray ? "yes" : "no");
    fflush(stdout);
#endif

    n = GET_LENGTH(obj);

    if(!isBuiltIn) {
        klass = ManagerFindClass(env, sig);
        get_Java_exception("ignoring exception", env);
    }

    if(ToJavaConverters && (which != JAVA_ANONYMOUS_REFERENCE)) { /* && klass != NULL_JAVA_OBJECT*/
        int done = 0;
        value = userLevelToJavaConversion(env, obj, klass, which, sig,
                                          n, isArray, &done);
        if(done)
            return value;
    }

#ifdef LDEBUG
    REprintf("No user level conversion\n");
#endif

    if(!isArray && n > 1 && isReference == FALSE) {
        PROBLEM "Length %ld > 1 but signature is (\"%s\")", n, sig
            ERROR;
        n = 1;
    }

    value = (void*) convertToJava(env, obj, isArray, which, n);

    if(actual_sig_ptr)
        *actual_sig_ptr = sig;

    return value;
}

/* #line 864 "RStoJava.cweb" */
void*
convertToJava(JNIEnv *env, USER_OBJECT_ obj, jboolean isArray, java_type whichType, long n)
{
    void* value = NULL_JAVA_OBJECT;
    long i;

#if LDEBUG
    printf("convertToJava %d %s %d\n", whichType, (isArray == JNI_TRUE? "yes" :"no"), (int) n);fflush(stdout);
    RS_dput(obj);
#endif

    switch(whichType) {
    case JAVA_DOUBLE: {
        double *in;
        jdouble *out;
        obj = AS_NUMERIC(obj);
        in = NUMERIC_POINTER(obj);
        if(sizeof(*in) != sizeof(*out)) {
            out = (jdouble *) Calloc(n, jdouble);
            for(i=0; i<n; i++)
                out[i] = in[i];
        }
        else
            out = in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        break;
    }
    case JAVA_FLOAT: {
        double *in;
        jfloat *out;
        obj = AS_NUMERIC(obj);
        in = NUMERIC_POINTER(obj);
        out = (jfloat *) Calloc(n, jfloat);
        for(i=0; i<n; i++)
            out[i] = in[i];
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        Free(out);
        break;
    }
    case JAVA_BOOLEAN: {
        RSInt *in; jboolean *out;
        obj = AS_LOGICAL(obj);
        in = LOGICAL_DATA(obj);
        if(sizeof(*in) != sizeof(*out)) {
            out = (jboolean *) Calloc(n, jfloat);
            for(i=0; i<n; i++)
                out[i] = in[i];
        }
        else
            out = (jboolean *)in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        break;
    }
    case JAVA_STRING: {
        jstring str;
        /*
          In S4, we can do the following. But R currently has the
          character vector not easily accessible as a char ** array.
          So we do things locally here.
          char **out;
          out = & CHAR_DEREF(CHARACTER_DATA(obj)[0]);
          value = toJavaType(env, whichType, n, (void *)obj, isArray);
        */
        if (isArray) {
            value = JavaStringArray(env, n);
            for(i=0; i < n; i++) {
                str = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(obj, i)));
                VMENV SetObjectArrayElement(env, value, i, str);
                VMENV DeleteLocalRef(env, str);
            }
        } else {
            if (n==1) {
                value = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(obj, 0)));
            } else {             /* n > 1 && !isArray caught by s_to_java_basic */
                value = VMENV NewStringUTF(env, "");
            }
        }
        break;
    }
    case JAVA_INT: {
        RSInt *in; jint *out;
        obj = AS_INTEGER(obj);
        in = INTEGER_POINTER(obj);
        if(sizeof(*in) != sizeof(*out)) {
            out = (jint *) Calloc(n, jint);
            for(i=0; i<n; i++) out[i] = in[i];
        }
        else
            out = (jint *)in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        break;
    }
    case JAVA_LONG: {
        RSInt *in; jlong *out;
        obj = AS_INTEGER(obj);
        in = INTEGER_POINTER(obj);
        if(sizeof(*in) != sizeof(*out)) {
            out = (jlong *) Calloc(n, jlong);
            for(i=0; i<n; i++) out[i] = in[i];
        }
        else out = (jlong *)in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        break;
    }
    case JAVA_SHORT: {
        RSInt *in; jshort *out;
        obj = AS_INTEGER(obj); in = INTEGER_POINTER(obj);
        if(sizeof(*in) != sizeof(*out)) {
            out = (jshort *) Calloc(n, jshort);
            for(i=0; i<n; i++) out[i] = in[i];
        }
        else out = (jshort *)in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
        break;
    }
    case JAVA_BYTE: {
#if 0
            /* FIXME: MTM: why is this commented out? */
            char *in; jbyte *out;
        obj = AS_RAW(obj);
        in = RAW_POINTER(obj);
        if(sizeof(*in) != sizeof(*out)) { /* this is not possible */
            out = (jbyte *) Calloc(n, jbyte);
            for(i=0; i<n; i++) out[i] = in[i];
        }
        else out = (jbyte *)in;
        value = toJavaType(env, whichType, n, (void *)out, isArray);
#endif
        break;
    }
    case JAVA_OBJECT: {
        /* this is a little different.  The strategy is to treat this as
           an uninterpreted recursive object, so we avoid using any S
           coercion methods -- these should be applied outside of the
           basic conversion code. */
#if 0
        USER_OBJECT_ *in;
        void **out;
        if(GET_MODE(obj) != VECSXP) /* must not use IS_RECURSIVE here */
            obj = AS_VECTOR(obj);
        in = RECURSIVE_DATA(obj);
        out = (void **) Calloc(n, void *);
        for(i=0; i<n; i++)
            out[i] = s_to_java_basic(env, in[i], NULL, NULL);
        value = toJavaType(env, whichType, n, (void *)out, isArray);
#endif

/* Now we assume this is a foreignReference. */
    }
        break;

    case JAVA_NAMED_REFERENCE:
    case JAVA_ANONYMOUS_REFERENCE:
        value = userReferenceToJava(env, obj, whichType);
        break;

    case RS_NAMED_REFERENCE:
    case RS_ANONYMOUS_REFERENCE:
#if LDEBUG
        REprintf("Creating Java object from R reference %d\n", whichType);
#endif
        value = NULL_JAVA_OBJECT;
        value = userReferenceToJava(env, obj, whichType);
        break;

    case JAVA_GENERIC_REFERENCE:
    {
        int i;
        jstring jid, jclassName, tmp;
        jobjectArray jtypes = NULL;

#if LDEBUG
        printf("Creating a JAVA_GENERIC_REFERENCE\n");
        RS_dput(obj);
#endif
        /* */
        if (!IS_LIST(obj)) {
            const char *rClass =
              CHAR_DEREF(STRING_ELT(GET_CLASS(obj), 0));
            const char *msg =
              create_message("%s: Object is a '%s', not a '%s'; %s",
                             "SJava::convertToJava", rClass,
                             "list", "missing converter?");
            throw_Java_exception(msg, "org/omegahat/R/Java/RException",
                                 env);
            return NULL_JAVA_OBJECT;
        }
        SEXP
          fid = VECTOR_ELT(obj, FOREIGN_REFERENCE_ID_SLOT),
          fname = VECTOR_ELT(obj, FOREIGN_REFERENCE_CLASSNAME_SLOT);
        jid = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(fid, 0)));
        jclassName =
          VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(fname, 0)));

        n = GET_LENGTH(VECTOR_ELT(obj, FOREIGN_REFERENCE_CLASS_SLOT));
        jtypes = JavaStringArray(env, n);
        for(i = 0 ; i < n ; i++) {
            SEXP fslot = VECTOR_ELT(obj, FOREIGN_REFERENCE_CLASS_SLOT);
            tmp = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(fslot, i)));
            VMENV SetObjectArrayElement(env, jtypes, i, tmp);
            VMENV DeleteLocalRef(env, tmp);

            tmp = VMENV GetObjectArrayElement(env, jtypes, i);
            {
#if LDEBUG
                const char *ptr;
                ptr = VMENV GetStringUTFChars(env, tmp, NULL);
                printf("+%d) %s\n",i,ptr );
                VMENV ReleaseStringUTFChars(env, tmp, ptr);
#endif
                VMENV DeleteLocalRef(env, tmp);
            }
        }
        value = RS(createForeignReference)(env, jid, jclassName, jtypes);
    }
    break;

    default:
    {
        char *sig = "";
        PROBLEM "\"%s\" is not a basic type with a known conversion from S to Java",
            sig ERROR;
        value = NULL_JAVA_OBJECT;
    }
    }

    return value;
}

/* #line 758 "RStoJava.cweb" */
USER_OBJECT_
s_create_default_vm()
{
    char *RS_NULL_STRING = (char*)NULL;
    const char *msg;
    USER_OBJECT_ ans = NEW_LOGICAL(1);

    msg = interface_start_VM(RS_NULL_STRING, RS_NULL_STRING, RS_NULL_STRING,
                             RS_NULL_STRING, (char **)NULL, 0);
    if(msg)
        PROBLEM "Couldn't start default Java VM: %s", msg
            ERROR;

    LOGICAL_POINTER(ans)[0] = 1;

    return ans;
}




/* #line 1214 "RStoJava.cweb" */
USER_OBJECT_
s_register_VM()
{
    JNIEnv *env;
    const char *msg;
    USER_OBJECT_ ans;
    jboolean wasAttached;
    ans = NEW_LOGICAL(1);
    env = getThreadJNIEnv(&wasAttached);
    if(env) {
        msg = do_native_setup(NULL, NULL, env);
        LOGICAL_DATA(ans)[0] = TRUE;
        REprintf("Attached ourselves to the JVM\n");
    }

    return ans;
}

#ifdef Unix
static void (*nop)(void);
#endif

void
s_start_VM(char **classPathEls, char **extraProperties, long *numExtraProperties,
           char **options, long *numOptions)
{
#ifdef REGISTER_NATIVES
    extern int RSJava_registerNatives(JNIEnv *);
#endif

    const char *msg;
    JavaVM *vmBuf;
    jsize nVMs = 0;
    if(JNI_GetCreatedJavaVMs(&vmBuf, 1, &nVMs) == 0 && nVMs > 0) {
        PROBLEM "%d Java VM(s) exist, ignoring initialization request", nVMs
            WARN;
        return;
    }

    msg = create_Java_vm(classPathEls[0], NULL, NULL,
                         extraProperties, *numExtraProperties,
                         /* options, (int) *numOptions,*/ 0, 1);

    if(msg) {
        PROBLEM
            "Could not start Java VM: %s", msg
            ERROR;
    }

    *numExtraProperties = TRUE;

    /* Now make certain the basic converters for the reflect meta objects
       and references are available.
    */
    RS_JAVA(registerDefaultConverters(getJNIEnv()));

#ifdef REGISTER_NATIVES
    RSJava_registerNatives(getJNIEnv());
#endif
}


void
registerJavaPolledEvents()
{
    /* Activate the polling events to query */
#ifdef Unix
    nop = R_PolledEvents;
    RS_lockEvaluator(NULL, JNI_TRUE, JNI_FALSE);
    R_PolledEvents = RS_JavaPolledEvents;
    R_wait_usec = 10000;
#endif
}




/* #line 106 "RStoJava.cweb" */
/* The order is important to allow the CHAR in the
   conversion switch statement be well-defined.
*/
/* #line 155 "RStoJava.cweb" */
#if !defined(_R_)
#if defined(CHAR)
#undef CHAR
#endif

#define CHAR(x) (x)
#define NULL_USER_OBJECT S_void
#endif




/* #line 232 "RStoJava.cweb" */
USER_OBJECT_
RS_genericJavaCall(USER_OBJECT_ qualifier, USER_OBJECT_ methodName,
                   USER_OBJECT_ userArgs,
                   USER_OBJECT_ userArgNames, USER_OBJECT_ returnValueName,
                   USER_OBJECT_ signatures, USER_OBJECT_ convert)
{
    jobject jqualifier, jval;
    jobjectArray jargs, jargNames;
    jstring
    jmethodName = NULL_JAVA_OBJECT,
    jreturnValueName = NULL_JAVA_OBJECT;
    jint jerr;

    USER_OBJECT_ user_ans;
    const char *returnName;
    int basicConvert;

    jboolean attached;
    JNIEnv *env = getThreadJNIEnv(&attached);
    if(env == NULL) {
        PROBLEM "Java VM not initialized; use .JavaInit()"
            ERROR;
    }

#if LDEBUG
    REprintf("Qualifier\n");
#endif
    if(GET_LENGTH(qualifier) > 0) {
        if(IS_CHARACTER(qualifier)) {
            /* convert directly to a Java string without going through the
               user-level conversion*/
            jqualifier = VMENV NewStringUTF(env, CHAR(STRING_ELT(qualifier, 0)));
        } else
            jqualifier = s_to_java_basic(env, qualifier, NULL, NULL);
    } else
        jqualifier = NULL_JAVA_OBJECT;

#if LDEBUG
    REprintf("Arguments\n");
#endif
    returnName  = (GET_LENGTH(returnValueName) > 0) ?
        CHAR_DEREF(STRING_ELT(returnValueName,0)) : "";
    jargs = argsToArray(env, userArgs, signatures);
    jargNames = argNamesToArray(env, userArgNames, FALSE);
    jmethodName =
    VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(methodName, 0)));
    jreturnValueName = VMENV NewStringUTF(env, returnName);

/*XXX need to release these at the end.  */

    basicConvert = JNI_TRUE;

    if(TYPEOF(convert) == LGLSXP && GET_LENGTH(convert) > 0)
        basicConvert = LOGICAL_DATA(convert)[0];

#if LDEBUG
    REprintf("calling generic omega method\n");
#endif
    jval = callGenericOmegaMethod(env, jqualifier, jmethodName,
                                  jargs, jargNames, jreturnValueName,
                                  basicConvert);

    CATCH_EXCEPTION;

#ifdef LDEBUG
    if(jval != NULL_JAVA_OBJECT) {
        REprintf("Return value class name %s\n",
                CHAR(getClassName(env, VMENV GetObjectClass(env, jval))));
    }
#endif

#if LDEBUG
    REprintf("converting answer\n");
#endif

    if(GET_LENGTH(convert) == 0) {
        user_ans = NULL_USER_OBJECT;
    } else {
        if(basicConvert == JNI_FALSE) {
            user_ans = anonymousAssign(env, jval, NULL);
        } else {
            user_ans = convertAnswer(env, returnValueName, jval, convert);
        }
    }

    CATCH_EXCEPTION;

    if (jval != NULL_JAVA_OBJECT)
        VMENV DeleteLocalRef(env, jval);
    VMENV DeleteLocalRef(env, jqualifier);
    VMENV DeleteLocalRef(env, jargs);
    VMENV DeleteLocalRef(env, jargNames);
    VMENV DeleteLocalRef(env, jreturnValueName);
    VMENV DeleteLocalRef(env, jmethodName);

    if ((jerr = releaseThreadJNIEnv(attached)) != JNI_OK) {
        PROBLEM "SJava internal: %s:%d",
            "RS_genericJavaCall:DetachCurrentThread", jerr
            ERROR;
    }

    return user_ans;
}



/* #line 530 "RStoJava.cweb" */
USER_OBJECT_
convertAnswer(JNIEnv *env, USER_OBJECT_ returnValueName, jobject val, USER_OBJECT_ converter)
{
    USER_OBJECT_ user_ans = NULL_USER_OBJECT;
    if(val == NULL_JAVA_OBJECT)
        return user_ans;

#ifdef LDEBUG
    REprintf("convertAnswer: have name for result? %d\n",
             (GET_LENGTH(returnValueName) == 0));
#endif


    if(TYPEOF(converter) == EXTPTRSXP ) {
        /* a routine */
        FromJavaConverter f;
        jclass type;

        if(R_ExternalPtrTag(converter) != Rf_install("native symbol")) {
            PROBLEM "Don't recognize the tag on the external pointer to use for conversion routine"
                ERROR;
        }

        f = (FromJavaConverter)  R_ExternalPtrAddr(converter);
        type = VMENV GetObjectClass(env, val);
        user_ans = f(val, type, env, FromJavaConverters);
        return user_ans;
    } else if(TYPEOF(converter) == CLOSXP) {
        /* Call the function */
        jclass type = VMENV GetObjectClass(env, val);
        user_ans = RS_JAVA(callRConverterFunction)(converter, val, type, env, NULL, ConverterFunction);
        return user_ans;
    } else if(TYPEOF(converter) == LGLSXP && !LOGICAL_DATA(converter)[0]) {
        user_ans = anonymousAssign(env, val, NULL);
        return user_ans;
    }


    if(GET_LENGTH(returnValueName) == 0) {
        /* convert the returned value. */

        if(val != NULL && val != NULL_JAVA_OBJECT) {
            const char *dataType;
            jstring className;
#ifdef LDEBUG
            REprintf("No return name. Getting class name\n");
#endif
            className = get_object_class_name(env, val);
            dataType = VMENV GetStringUTFChars(env, className, NULL);
            user_ans = s_from_java_basic(env, val, dataType, TRUE, TRUE);
            VMENV ReleaseStringUTFChars(env, className, dataType);
            VMENV DeleteLocalRef(env, className);
        }
    } else {
        user_ans = s_from_java_basic(env, val, NULL, TRUE, TRUE);
    }

    return user_ans;
}



/* #line 604 "RStoJava.cweb" */
jobjectArray
argsToArray(JNIEnv *env, USER_OBJECT_ userArgs, USER_OBJECT_ signatures)
{
    const char *actual_sig;
    long numSignatures = GET_LENGTH(signatures);
    long numArgs = GET_LENGTH(userArgs);
    int i;
    jobjectArray argArray;

    /* First, we traverse the list of arguments and convert them to
       Java objects.  We gather these into an Object[] array.
    */
    argArray = JavaObjectArray(env, numArgs);
    for(i = 0; i < numArgs; i++) {
        jobject tmp;

        if(VECTOR_ELT(userArgs,i) == NULL_USER_OBJECT)
            continue;

        tmp = (jobject)
          s_to_java_basic(env, VECTOR_ELT(userArgs, i),
                          i < numSignatures ?
                          CHAR(STRING_ELT(signatures, i)) : NULL,
                          &actual_sig);
        /* now put this into an Object array. */
        VMENV SetObjectArrayElement(env, argArray, i, tmp);
    }

    return argArray;
}

/* #line 646 "RStoJava.cweb" */
jobjectArray
argNamesToArray(JNIEnv *env, USER_OBJECT_ userArgNames, int addExtraValue)
{
    jobjectArray argNames = NULL_JAVA_OBJECT;
    long numNames = GET_LENGTH(userArgNames);


    if(numNames > 0) {

#if 1
/* For the moment, we just write the code here rather than using
   the toJavaType(). This latter routine is now fixed and in the case
   of a JAVA_STRING array expects the pointer to the actual character() vector.
*/
        jobjectArray array;
        jclass klass = VMENV FindClass(env, "java/lang/String");
        int i;
        jstring str;
        array = VMENV NewObjectArray(env, numNames, klass, NULL);
        for(i = 0; i < numNames; i++) {
            str = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(userArgNames, i)));
            VMENV SetObjectArrayElement(env, array, i, str);
            VMENV DeleteLocalRef(env, str);
        }

        argNames = array;
#else
        /* Now works for R by passing in the top-level character object */
        argNames =
            toJavaType(env, JAVA_STRING, numNames,
                       (void*)userArgNames, (jboolean) JNI_TRUE);
#endif
    }

    return argNames;
}


/* #line 690 "RStoJava.cweb" */
jobjectArray
JavaObjectArray(JNIEnv *env, int length)
{
    /* Should be a pthread_once */
    static jclass clazz = NULL;
    jobjectArray arr;

    if(clazz == NULL) {
        clazz = VMENV FindClass(env, "java/lang/Object");
        clazz = VMENV NewGlobalRef(env, clazz);
    }


    arr = VMENV NewObjectArray(env, (jint) length, clazz, NULL);

    return arr;
}

/* #line 723 "RStoJava.cweb" */
jobjectArray
JavaStringArray(JNIEnv *env, int length)
{
    /* Should be a pthread_once */
    static jclass clazz = NULL;
    jobjectArray arr;

    if(clazz == NULL) {
        clazz = VMENV FindClass(env, "java/lang/String");
        clazz = VMENV NewGlobalRef(env, clazz);
    }

    arr = VMENV NewObjectArray(env, (jint) length, clazz, NULL);

    return arr;
}


/* #line 385 "RStoJava.cweb" */
USER_OBJECT_
RS_ConstructorJavaCall(USER_OBJECT_ className, USER_OBJECT_ userArgs,
                       USER_OBJECT_ userArgNames, USER_OBJECT_ returnValueName,
                       USER_OBJECT_ signatures, USER_OBJECT_ convert)
{
    JNIEnv *env = getJNIEnv();

    jstring jreturnValueName = NULL_JAVA_OBJECT;
    jobject val;
    USER_OBJECT_ user_ans;
    jstring JclassName;

    jobjectArray args;
    jobjectArray argNames;
    jboolean basicConvert = JNI_TRUE;

    if(env == NULL) {
        PROBLEM "Java VM not available, use .JavaInit()"
            ERROR;
    }

#ifdef LDEBUG
    REprintf("RS_ConstructorJavaCall: env = %p\n", env);
#endif

    args = argsToArray(env, userArgs, signatures);
    argNames = argNamesToArray(env, userArgNames, FALSE);
    JclassName =
        VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(className, 0)));

    if(GET_LENGTH(returnValueName) > 0) {
        jreturnValueName =
            VMENV NewStringUTF(env,
                               CHAR_DEREF(STRING_ELT(returnValueName, 0)));
    } else
        jreturnValueName = VMENV NewStringUTF(env, "");

    if(TYPEOF(convert) == LGLSXP && LENGTH(convert))
        basicConvert = LOGICAL_DATA(convert)[0];

    val = callConstructorMethod(env, JclassName, args, argNames,
                                jreturnValueName, basicConvert);

    VMENV DeleteLocalRef(env, jreturnValueName);
    VMENV DeleteLocalRef(env, JclassName);

    CATCH_EXCEPTION;

    if(basicConvert == JNI_FALSE) {
        user_ans = anonymousAssign(env, val, NULL);
    } else {
        user_ans =  convertAnswer(env, returnValueName, val, convert);
    }
    VMENV DeleteLocalRef(env, val);

    CATCH_EXCEPTION;

    return user_ans;
}


/* #line 458 "RStoJava.cweb" */
USER_OBJECT_
RS_OmegahatExpression(USER_OBJECT_ expressionString,
                      USER_OBJECT_ userArgs, USER_OBJECT_ userArgNames,
                      USER_OBJECT_ returnValueName,
                      USER_OBJECT_ signatures,
                      USER_OBJECT_ convert)
{
    JNIEnv *env = getJNIEnv();

    jstring jreturnValueName = NULL_JAVA_OBJECT;
    jstring expression;
    jobject val;
    USER_OBJECT_ user_ans = NULL_USER_OBJECT;

    jobjectArray args = argsToArray(env, userArgs, signatures);
    jobjectArray argNames = argNamesToArray(env, userArgNames, FALSE);
    jboolean basicConvert = JNI_TRUE;

    if(TYPEOF(convert) == LGLSXP && LENGTH(convert))
        basicConvert = LOGICAL_DATA(convert)[0];

    if(GET_LENGTH(returnValueName) > 0) {
        jreturnValueName = VMENV NewStringUTF(env, CHAR_DEREF(STRING_ELT(returnValueName, 0)));
    }


    expression = VMENV NewStringUTF(env,  CHAR_DEREF(STRING_ELT(expressionString, 0)));
    val = evalOmegaExpression(env, expression, args, argNames, jreturnValueName, basicConvert);

    if(basicConvert == JNI_FALSE) {
        user_ans = anonymousAssign(env, val, NULL);
    } else {
        user_ans =  convertAnswer(env, returnValueName, val, convert);
    }

    return user_ans;
}



/* #line 747 "RStoJava.cweb" */
JNIEnv *SgetJNIEnv()
{
    if(!exists_Java_vm())
        s_create_default_vm();

    return getJNIEnv();
}



/* #line 1133 "RStoJava.cweb" */
USER_OBJECT_
javaToUserString(JNIEnv *env, jobject val)
{
    const char *inString;
    USER_OBJECT_ ans;
    inString = VMENV GetStringUTFChars(env, (jstring) val, NULL);
    PROTECT(ans = NEW_CHARACTER(1));
    SET_STRING_ELT(ans, 0, COPY_TO_USER_STRING((char *)inString));
    VMENV ReleaseStringUTFChars(env, (jstring) val, inString);
    UNPROTECT(1);

    return ans;
}


/* #line 1162 "RStoJava.cweb" */
jclass
ManagerFindClass(JNIEnv *env, const char *className)
{
    jclass klass;
    jstring jstr;
    jmethodID findClassID;
    jobject mgr = getInterfaceManager();
    jstr = VMENV NewStringUTF(env, className);
    klass = VMENV GetObjectClass(env, mgr);
    findClassID = VMENV GetMethodID(env, klass, "findClass", "(Ljava/lang/String;)Ljava/lang/Class;");
    klass =  VMENV CallObjectMethod(env, mgr, findClassID, jstr);

    return klass;
}



/* #line 1184 "RStoJava.cweb" */
void
RS_JAVA(isJVMInitialized)(long *wellThen)
{
    JNIEnv *env = getJNIEnv();
    *wellThen = (env != NULL);
}



/* #line 1587 "RStoJava.cweb" */
const char  * const BaseOmegahatReferenceClass = "OmegahatReference";
const char  * const NamedOmegahatReferenceClass = "NamedOmegahatReference";
const char  * const AnonymousOmegahatReferenceClass = "AnonymousOmegahatReference";



/* #line 1299 "RStoJava.cweb" */
/* catch exception, if any; When S is in charge turn an exception into
   an S error.  Otherwise return TRUE if there is an exception.   If
   error==FALSE, clear the exception
*/

#define IS_S_ENVIRONMENT (1 || env == getJNIEnv())

boolean
s_catch_exception(JNIEnv *env, boolean asError)
{
    char *msg;
    jthrowable exception;
    if(!env)
        env = interface_VM_env();

    exception = VMENV ExceptionOccurred(env);

    if(!exception)
        return FALSE;

    exception = VMENV NewGlobalRef(env, exception);

    /* Probably need to make a reference here. */
    VMENV ExceptionClear(env);

#if 1
    /* Assign this to the Omegahat database as .Last.exception.  */
    assignException(env, exception);
#endif


    if(asError) {
        throwRJavaError(env, exception);
        PROBLEM "Exception occurred in Java"
            ERROR;
    } else {
        PROBLEM "Exception occurred in Java"
            WARN;
    }



    msg = (char *) get_Java_exception(NULL, env);
    if(msg == NULL)
        return FALSE;
    if(IS_S_ENVIRONMENT) {
        /* prepare an S error message before clearing Java exception. */
#if 1
        const char *name;
        name = get_Java_exception_class(env);
        PROBLEM "Java exception(%s): \"%s\"", name, msg
            END_MESSAGE;
        msg = GET_MESSAGE;
#else
        msg = "Error in Java";
#endif

        clear_Java_exception(env);
        if(asError) {
            PROBLEM "%s", msg
                ERROR;
        } else {
            PROBLEM "Clearing %s", msg WARN;
        }
    } else {

/*    if(!error) */
        clear_Java_exception(env);
        return TRUE;
    }

    return FALSE;
}



/* #line 1381 "RStoJava.cweb" */
void
assignException(JNIEnv *env, jthrowable obj)
{
    jobject mgr = getInterfaceManager();
    jmethodID mid;
    jclass klass;

    klass = VMENV GetObjectClass(env, mgr);
    mid = VMENV GetMethodID(env, klass, "assignException", "(Ljava/lang/Throwable;)V");
    if(!mid) {
        REprintf("Can't find assignException method in interface manager %p %p\n", mgr, klass);
        return;
    }

    VMENV CallVoidMethod(env, mgr, mid, obj);
}

/* #line 1404 "RStoJava.cweb" */
void
throwRJavaError(JNIEnv *env, jthrowable exc)
{
    USER_OBJECT_ rerror;
    USER_OBJECT_ e, klass;

    PROTECT(rerror = anonymousAssign(env, exc, NULL));  /*, R_MakeExternalPtr(exc, Rf_install("JavaException"), R_NilValue)); */
    PROTECT(klass = NEW_CHARACTER(4));
    SET_STRING_ELT(klass, 0, COPY_TO_USER_STRING("JavaException"));
    SET_STRING_ELT(klass, 1, COPY_TO_USER_STRING("condition"));
    SET_STRING_ELT(klass, 2, COPY_TO_USER_STRING("AnonymousOmegahatReference"));
    SET_STRING_ELT(klass, 3, COPY_TO_USER_STRING("OmegahatReference"));

    SET_CLASS(rerror, klass);

    PROTECT(e = allocVector(LANGSXP, 2));
    SETCAR(e, Rf_install("stop"));
    SETCAR(CDR(e), rerror);

    Rf_eval(e, R_GlobalEnv);

    UNPROTECT(3);
    return;

}

void
RS_dput(USER_OBJECT_ obj)
{
    USER_OBJECT_ call;
    PROTECT(call = allocVector(LANGSXP, 2));
    SETCAR(call, Rf_install("dput"));
    SETCAR(CDR(call), obj);
    Rf_eval(call, R_GlobalEnv);
    UNPROTECT(1);
}

/* #line 1447 "RStoJava.cweb" */
USER_OBJECT_
asOmegaReferenceUserObject(JNIEnv *env, jobject omegaObject, FuzzyBoolean isArray)
{
    USER_OBJECT_ ans = NULL_USER_OBJECT;
    void *val;
    char *SclassName = NULL;
    jstring className;

    if(omegaObject == NullObject) {
#ifdef LDEBUG
        REprintf("[asOmegaReferenceUserObject] Java object is null\n");
#endif
        return ans;
    }

    if(OmegaNamedReferenceClass  == NULL) {
        if(initReferenceClasses(env) < 1) {
#ifdef LDEBUG
            REprintf("Reference classes are not initialized\n");
#endif
            return ans; /* Big problem! */
        }
    }

    if( VMENV IsInstanceOf(env, omegaObject, OmegaNamedReferenceClass) == JNI_TRUE) {
        SclassName = "NamedOmegahatReference";
    } else  if( VMENV IsInstanceOf(env, omegaObject, OmegaAnonymousReferenceClass) == JNI_TRUE) {
        SclassName = "AnonymousOmegahatReference";
    } else {
        /* If we wanted to report this as an error,
           then we would compute the name of the class of the
           object in question. However, s_from_java_basic calls this
           when it is attempting to find a suitable converter. So we would
           have to add an argument to this routine to indicate whether it
           is an error or not to get to this point.
        */
#if 0
        jstring tmp = get_object_class_name(env, omegaObject);
        char* klassName = VMENV GetStringUTFChars(env, tmp, NULL);
        REprintf("[asOmegaReferenceUserObject] Java object is not a reference. Class is %s\n", klassName);
        VMENV ReleaseStringUTFChars(env, tmp, klassName);
#endif
        return ans;
    }

    /* Get the key/name of the object in the database. */
    val = VMENV CallObjectMethodA(env, omegaObject, ReferenceNameMethodID, NULL);
    /* Get the name of the class of the real object. */
    className = VMENV CallObjectMethodA(env, omegaObject, ReferenceClassNameMethodID, NULL);

    if(isArray == FUZZY_MAYBE)
        isArray = isJavaObjectArray(env, omegaObject, className);

    ans =  createUserReferenceObject(env, (const char*)SclassName, (jstring) val, className, isArray);

    return ans;
}

/* #line 1548 "RStoJava.cweb" */
USER_OBJECT_
createUserReferenceObject(JNIEnv *env, const char *SclassName, jstring value, jstring realClassName, int isArray)
{
    USER_OBJECT_ ans = NULL_USER_OBJECT;
    USER_OBJECT_ className, slotNames;

    PROTECT(ans = NEW_LIST(2));

    /* Set the `name' slot, which we happen to know is slot 0! */
    SET_VECTOR_ELT(ans, 0, javaToUserString(env, value));
    SET_VECTOR_ELT(ans, 1, javaToUserString(env, realClassName));

    PROTECT(slotNames = NEW_CHARACTER(2));
    SET_STRING_ELT(slotNames, 0, COPY_TO_USER_STRING("key"));
    SET_STRING_ELT(slotNames, 1, COPY_TO_USER_STRING("className"));
    SET_NAMES(ans, slotNames);
    UNPROTECT(1); /* slotNames */

    PROTECT(className = NEW_CHARACTER(isArray ? 3 : 2));
    if(isArray)
        SET_STRING_ELT(className, 0, COPY_TO_USER_STRING("JavaArrayReference"));

    SET_STRING_ELT(className, isArray ? 1 : 0, COPY_TO_USER_STRING((char*) SclassName));
    SET_STRING_ELT(className, isArray ? 2 : 1, COPY_TO_USER_STRING("OmegahatReference"));
    SET_CLASS(ans, className);

    UNPROTECT(2); /* className & ans*/

    return ans;
}

/* #line 1617 "RStoJava.cweb" */
jobject
userReferenceToJava(JNIEnv *env, USER_OBJECT_ obj, java_type which)
{
    const char *name;
    jobject jobj;

#if 0
    if(R_IS(obj, BaseJavaReferenceClass) == FALSE) {
        PROBLEM "Cannot convert object to OmegahatReference"
            ERROR;
    }
#endif
    /* Get the ID part. */
    if (!IS_LIST(obj)) {
      const char *rClass =
        CHAR_DEREF(STRING_ELT(GET_CLASS(obj), 0));
      const char *msg =
        create_message("%s: object is a '%s', not a '%s'; %s",
                       "SJava::userReferenceToJava", rClass,
                       "list", "missing converter?");
      throw_Java_exception(msg, "org/omegahat/R/Java/RException",
                           env);
      return NULL_JAVA_OBJECT;
    }
    name = CHAR_DEREF(STRING_ELT(VECTOR_ELT(obj, 0), 0));
#if LDEBUG
    REprintf("[userReferenceToJava] %s %d\n", name, which);
    fflush(stdout);
#endif

    if(which == RS_NAMED_REFERENCE || which == RS_ANONYMOUS_REFERENCE) {
        jclass klass;
        jmethodID mid;
        jstring jname;
        klass = RS_JAVA(getForeignReferenceClass)(obj,  env);
        jname = VMENV NewStringUTF(env, name);
        mid = VMENV GetMethodID(env, klass, "<init>", "(Ljava/lang/String;)V");
        jobj = VMENV NewObject(env, klass, mid, jname);
        VMENV DeleteLocalRef(env, jname);
    } else {
        jobj = createOmegaReference(env, name, NULL, which);
    }

    return jobj;
}

/* #line 1659 "RStoJava.cweb" */
jclass RS_JAVA(getForeignReferenceClass)(USER_OBJECT_ obj, JNIEnv *env)
{
    char *className;
    jclass klass;
#if defined(_R_)
    className = "org/omegahat/R/Java/RForeignReference";
#else
    className = "org/omegahat/S/Java/SForeignReference";
#endif
    klass = VMENV FindClass(env, className);

    return klass;
}


/* #line 1680 "RStoJava.cweb" */
int
R_IS(USER_OBJECT_ obj, const char * const name)
{
    int i, n;
    USER_OBJECT_ className;

    className = GET_CLASS(obj);
    n = GET_LENGTH(className);

    for(i=0;i<n;i++) {
        if(strcmp(name, CHAR_DEREF(STRING_ELT(className, i))) == 0) {
#if LDEBUG
            REprintf("R_IS: %s\n", name);
            RS_dput(obj);
#endif
            return TRUE;
        }
    }
    return FALSE;
}



/* #line 1513 "RStoJava.cweb" */
FuzzyBoolean
isJavaObjectArray(JNIEnv *env, jobject obj, jstring className)
{
    const char *str;
    int ans;

    str = VMENV GetStringUTFChars(env, className, NULL);
    ans = (str[0] == '[');
    VMENV ReleaseStringUTFChars(env, className, str);

    return ans == JNI_TRUE ? FUZZY_TRUE : FUZZY_FALSE;
}


/* #line 339 "RStoJava.cweb" */
USER_OBJECT_
R_getJavaBoolean(jobject obj, jclass type, JNIEnv *env, RSFromJavaConverter *converters)
{
    jboolean b;
    const char *msg;
    b = VMENV CallBooleanMethod(env, obj, BasicClassValueAccessors[JAVA_BOOLEAN]);
    if((msg = get_Java_exception(NULL, env))) {
        PROBLEM "can't extract boolean value."
            ERROR;
    }
    return ScalarLogical(b);
}




/* #line 1709 "RStoJava.cweb" */
JNIEXPORT jboolean JNICALL
Java_org_omegahat_R_Java_REvaluator_lockEvaluator(JNIEnv *env, jboolean on)
{
    return RS_lockEvaluator(env, on, JNI_FALSE);
}

jboolean
RS_lockEvaluator(JNIEnv *env, jboolean on, jboolean announce)
{
    jint status;
    jobject lockObj;

//XXX
    announce = JNI_FALSE;

    if(!env) {
        env = getJNIEnv();
        if(!env)
            return JNI_FALSE;
    }

    lockObj = RS_getEvaluatorLock(env);

    if(!lockObj || lockObj == NULL_JAVA_OBJECT)
        return JNI_FALSE;

    if(on) {
        status = VMENV MonitorEnter(env, lockObj);

        if(announce)
            REprintf("Locked evaluator\n");
    } else {
        status = VMENV MonitorExit(env, lockObj);
        if(announce)
            REprintf("unlocking evaluator\n");
    }


    if(status != JNI_OK) {
        CATCH_EXCEPTION;
    }

#ifdef LDEBUG
    REprintf("Completed RS_lockEvaluator %s\n", on ? "lock" : "unlock");f
#endif

    return status == JNI_OK ? JNI_TRUE : JNI_FALSE;
}

/* #line 1808 "RStoJava.cweb" */
jobject
RS_getEvaluatorLock(JNIEnv *env)
{
    static jfieldID fid = 0;
    jclass klass;
    jobject ans;

    if(VMENV ExceptionCheck(env)) {
        REprintf("Clearing exception before getting evaluator lock\n");
        RJava_printStackTrace(env);
        VMENV ExceptionClear(env);
    }

    klass = VMENV FindClass(env, "org/omegahat/R/Java/REvaluator");

    if(VMENV ExceptionCheck(env)) {
        REprintf("[R_getEvaluatorLock] Couldn't load REvaluator class\n");
        VMENV ExceptionDescribe(env);
#ifdef Unix
        R_PolledEvents = nop;
#endif
        return NULL_JAVA_OBJECT;
    }
    fid = VMENV GetStaticFieldID(env, klass, "Monitor", "Ljava/lang/String;");

    get_Java_exception("", env);

    ans = VMENV GetStaticObjectField(env, klass, fid);

    return ans;
}


/* #line 1790 "RStoJava.cweb" */
void
RS_yieldEvaluator(JNIEnv *env)
{
    jmethodID mid;
    jclass klass;

    klass = VMENV FindClass(env, "org/omegahat/R/Java/REvaluator");
    mid = VMENV GetStaticMethodID(env, klass, "yieldThread", "()V");
    VMENV CallStaticVoidMethod(env, klass, mid);
}


/* #line 1764 "RStoJava.cweb" */
void RS_JavaPolledEvents(void)
{
    JNIEnv *env;

    env = getJNIEnv();

    if(!env)
        return;


    VMENV ExceptionClear(env);

    if(RS_lockEvaluator(env, JNI_FALSE, JNI_FALSE) != JNI_TRUE)
        return;

    RS_yieldEvaluator(env);

    RS_lockEvaluator(env, JNI_TRUE, JNI_FALSE);
}


/* #line 1862 "RStoJava.cweb" */
/*
  Copyright (c) 1998, 1999 The Omega Project for Statistical Computing.
  All rights reserved.

*/
/*
  GNU GENERAL PUBLIC LICENSE

  Version 2, June 1991

  Copyright (C) 1989, 1991 Free Software Foundation, Inc. 675 Mass
  Ave, Cambridge, MA 02139, USA. Everyone is permitted to copy and
  distribute verbatim copies of this license document, but changing it
  is not allowed.

  Preamble

  The licenses for most software are designed to take away your
  freedom to share and change it. By contrast, the GNU General Public
  License is intended to guarantee your freedom to share and change
  free software--to make sure the software is free for all its users.
  This General Public License applies to most of the Free Software
  Foundation's software and to any other program whose authors commit
  to using it. (Some other Free Software Foundation software is
  covered by the GNU Library General Public License instead.) You can
  apply it to your programs, too.

  When we speak of free software, we are referring to freedom, not
  price. Our General Public Licenses are designed to make sure that
  you have the freedom to distribute copies of free software (and
  charge for this service if you wish), that you receive source code
  or can get it if you want it, that you can change the software or
  use pieces of it in new free programs; and that you know you can do
  these things.

  To protect your rights, we need to make restrictions that forbid
  anyone to deny you these rights or to ask you to surrender the
  rights. These restrictions translate to certain responsibilities for
  you if you distribute copies of the software, or if you modify it.

  For example, if you distribute copies of such a program, whether
  gratis or for a fee, you must give the recipients all the rights
  that you have. You must make sure that they, too, receive or can get
  the source code. And you must show them these terms so they know
  their rights.

  We protect your rights with two steps: (1) copyright the software,
  and (2) offer you this license which gives you legal permission to
  copy, distribute and/or modify the software.

  Also, for each author's protection and ours, we want to make certain
  that everyone understands that there is no warranty for this free
  software. If the software is modified by someone else and passed on,
  we want its recipients to know that what they have is not the
  original, so that any problems introduced by others will not reflect
  on the original authors' reputations.

  Finally, any free program is threatened constantly by software
  patents. We wish to avoid the danger that redistributors of a free
  program will individually obtain patent licenses, in effect making
  the program proprietary. To prevent this, we have made it clear that
  any patent must be licensed for everyone's free use or not licensed
  at all.

  The precise terms and conditions for copying, distribution and
  modification follow.

  GNU GENERAL PUBLIC LICENSE
  TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
  a notice placed by the copyright holder saying it may be distributed
  under the terms of this General Public License. The "Program",
  below, refers to any such program or work, and a "work based on the
  Program" means either the Program or any derivative work under
  copyright law: that is to say, a work containing the Program or a
  portion of it, either verbatim or with modifications and/or
  translated into another language. (Hereinafter, translation is
  included without limitation in the term "modification".) Each
  licensee is addressed as "you".

  Activities other than copying, distribution and modification are not
  covered by this License; they are outside its scope. The act of
  running the Program is not restricted, and the output from the
  Program is covered only if its contents constitute a work based on
  the Program (independent of having been made by running the
  Program). Whether that is true depends on what the Program does.

  1. You may copy and distribute verbatim copies of the Program's
  source code as you receive it, in any medium, provided that you
  conspicuously and appropriately publish on each copy an appropriate
  copyright notice and disclaimer of warranty; keep intact all the
  notices that refer to this License and to the absence of any
  warranty; and give any other recipients of the Program a copy of
  this License along with the Program.

  You may charge a fee for the physical act of transferring a copy,
  and you may at your option offer warranty protection in exchange for
  a fee.

  2. You may modify your copy or copies of the Program or any portion
  of it, thus forming a work based on the Program, and copy and
  distribute such modifications or work under the terms of Section 1
  above, provided that you also meet all of these conditions:

  a) You must cause the modified files to carry prominent notices
  stating that you changed the files and the date of any change.

  b) You must cause any work that you distribute or publish, that in
  whole or in part contains or is derived from the Program or any part
  thereof, to be licensed as a whole at no charge to all third parties
  under the terms of this License.

  c) If the modified program normally reads commands interactively
  when run, you must cause it, when started running for such
  interactive use in the most ordinary way, to print or display an
  announcement including an appropriate copyright notice and a notice
  that there is no warranty (or else, saying that you provide a
  warranty) and that users may redistribute the program under these
  conditions, and telling the user how to view a copy of this License.
  (Exception: if the Program itself is interactive but does not
  normally print such an announcement, your work based on the Program
  is not required to print an announcement.)

  These requirements apply to the modified work as a whole. If
  identifiable sections of that work are not derived from the Program,
  and can be reasonably considered independent and separate works in
  themselves, then this License, and its terms, do not apply to those
  sections when you distribute them as separate works. But when you
  distribute the same sections as part of a whole which is a work
  based on the Program, the distribution of the whole must be on the
  terms of this License, whose permissions for other licensees extend
  to the entire whole, and thus to each and every part regardless of
  who wrote it.

  Thus, it is not the intent of this section to claim rights or
  contest your rights to work written entirely by you; rather, the
  intent is to exercise the right to control the distribution of
  derivative or collective works based on the Program.

  In addition, mere aggregation of another work not based on the
  Program with the Program (or with a work based on the Program) on a
  volume of a storage or distribution medium does not bring the other
  work under the scope of this License.

  3. You may copy and distribute the Program (or a work based on it,
  under Section 2) in object code or executable form under the terms
  of Sections 1 and 2 above provided that you also do one of the
  following:

  a) Accompany it with the complete corresponding machine-readable
  source code, which must be distributed under the terms of Sections 1
  and 2 above on a medium customarily used for software interchange;
  or,

  b) Accompany it with a written offer, valid for at least three
  years, to give any third party, for a charge no more than your cost
  of physically performing source distribution, a complete
  machine-readable copy of the corresponding source code, to be
  distributed under the terms of Sections 1 and 2 above on a medium
  customarily used for software interchange; or,

  c) Accompany it with the information you received as to the offer to
  distribute corresponding source code. (This alternative is allowed
  only for noncommercial distribution and only if you received the
  program in object code or executable form with such an offer, in
  accord with Subsection b above.)

  The source code for a work means the preferred form of the work for
  making modifications to it. For an executable work, complete source
  code means all the source code for all modules it contains, plus any
  associated interface definition files, plus the scripts used to
  control compilation and installation of the executable. However, as
  a special exception, the source code distributed need not include
  anything that is normally distributed (in either source or binary
  form) with the major components (compiler, kernel, and so on) of the
  operating system on which the executable runs, unless that component
  itself accompanies the executable.

  If distribution of executable or object code is made by offering
  access to copy from a designated place, then offering equivalent
  access to copy the source code from the same place counts as
  distribution of the source code, even though third parties are not
  compelled to copy the source along with the object code.

  4. You may not copy, modify, sublicense, or distribute the Program
  except as expressly provided under this License. Any attempt
  otherwise to copy, modify, sublicense or distribute the Program is
  void, and will automatically terminate your rights under this
  License. However, parties who have received copies, or rights, from
  you under this License will not have their licenses terminated so
  long as such parties remain in full compliance.

  5. You are not required to accept this License, since you have not
  signed it. However, nothing else grants you permission to modify or
  distribute the Program or its derivative works. These actions are
  prohibited by law if you do not accept this License. Therefore, by
  modifying or distributing the Program (or any work based on the
  Program), you indicate your acceptance of this License to do so, and
  all its terms and conditions for copying, distributing or modifying
  the Program or works based on it.

  6. Each time you redistribute the Program (or any work based on the
  Program), the recipient automatically receives a license from the
  original licensor to copy, distribute or modify the Program subject
  to these terms and conditions. You may not impose any further
  restrictions on the recipients' exercise of the rights granted
  herein. You are not responsible for enforcing compliance by third
  parties to this License.

  7. If, as a consequence of a court judgment or allegation of patent
  infringement or for any other reason (not limited to patent issues),
  conditions are imposed on you (whether by court order, agreement or
  otherwise) that contradict the conditions of this License, they do
  not excuse you from the conditions of this License. If you cannot
  distribute so as to satisfy simultaneously your obligations under
  this License and any other pertinent obligations, then as a
  consequence you may not distribute the Program at all. For example,
  if a patent license would not permit royalty-free redistribution of
  the Program by all those who receive copies directly or indirectly
  through you, then the only way you could satisfy both it and this
  License would be to refrain entirely from distribution of the
  Program.

  If any portion of this section is held invalid or unenforceable
  under any particular circumstance, the balance of the section is
  intended to apply and the section as a whole is intended to apply in
  other circumstances.

  It is not the purpose of this section to induce you to infringe any
  patents or other property right claims or to contest validity of any
  such claims; this section has the sole purpose of protecting the
  integrity of the free software distribution system, which is
  implemented by public license practices. Many people have made
  generous contributions to the wide range of software distributed
  through that system in reliance on consistent application of that
  system; it is up to the author/donor to decide if he or she is
  willing to distribute software through any other system and a
  licensee cannot impose that choice.

  This section is intended to make thoroughly clear what is believed
  to be a consequence of the rest of this License.

  8. If the distribution and/or use of the Program is restricted in
  certain countries either by patents or by copyrighted interfaces,
  the original copyright holder who places the Program under this
  License may add an explicit geographical distribution limitation
  excluding those countries, so that distribution is permitted only in
  or among countries not thus excluded. In such case, this License
  incorporates the limitation as if written in the body of this
  License.

  9. The Free Software Foundation may publish revised and/or new
  versions of the General Public License from time to time. Such new
  versions will be similar in spirit to the present version, but may
  differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
  Program specifies a version number of this License which applies to
  it and "any later version", you have the option of following the
  terms and conditions either of that version or of any later version
  published by the Free Software Foundation. If the Program does not
  specify a version number of this License, you may choose any version
  ever published by the Free Software Foundation.

  10. If you wish to incorporate parts of the Program into other free
  programs whose distribution conditions are different, write to the
  author to ask for permission. For software which is copyrighted by
  the Free Software Foundation, write to the Free Software Foundation;
  we sometimes make exceptions for this. Our decision will be guided
  by the two goals of preserving the free status of all derivatives of
  our free software and of promoting the sharing and reuse of software
  generally.

  NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO
  WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.
  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR
  OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY
  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
  PARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND
  PERFORMANCE OF THE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE
  DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR
  CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
  WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY
  AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU
  FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR
  CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE
  PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
  RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
  FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF
  SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGES.

  END OF TERMS AND CONDITIONS
*/

